// 32colors by gynvael.coldwind//vx
// Inspired by the old Perihelion game for Amiga.
// Thanks to Arashi Coldwind for suggesting using hue-based distance
// instead of rgb-based distance.
// And thanks to Unavowed 
//
// To compiled:
//   g++ 32colors.cpp -Wall -Wextra -lSDL -lSDL_image -o 32colors.exe
//
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdint.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>

// Because I don't like SDL main redirecting my stdout.
#ifdef _WIN32
#  undef main
#endif

// Structures
struct RGBf {
  float r,g,b;
};

#pragma pack(push,1)
struct RGBu {
  uint8_t r,g,b;
};
#pragma pack(pop)

struct IMGf {
  int w,h;
  RGBf *data;
};

// Globals
float g_Angle;
bool  g_UseHueDist;

#define FF(a) ((float)(a) / 255.0f)

RGBf g_Sepia = { FF(156), FF(70), FF(20) };
RGBf g_SepiaNorm = g_Sepia;

RGBf g_Gray  = { FF(128+32), FF(128+16), FF(128-16) };
RGBf g_GrayNorm = g_Gray;

// I really like Bayer's dithering. Maybe it's because it was used
// on Windows 3.X (i.e. in the good old times ;>).
// http://en.wikipedia.org/wiki/Ordered_dithering
const float g_BayerMatrix[4][4] = {
  { FF(1),  FF(9),  FF(3),  FF(11) },
  { FF(13), FF(5),  FF(15), FF(7)  },
  { FF(4),  FF(12), FF(2),  FF(10) },
  { FF(16), FF(8),  FF(14), FF(6)  }
};

// Convert SDL surface to my float-driven image struct.
IMGf *SurfaceToIMGf(SDL_Surface *s)
{
  IMGf *img = new IMGf;
  img->w = s->w;
  img->h = s->h;
  img->data = new RGBf[img->w * img->h];
  RGBu *p = (RGBu*)s->pixels;
  RGBf *d = img->data;

  for(int j = 0; j < s->h; j++) {
    for(int i = 0; i < s->w; i++, p++,d++) {
      d->r = p->r / 255.0f;
      d->g = p->g / 255.0f;
      d->b = p->b / 255.0f;
    }
  }

  return img;
}

// The other way arround (I don't want to allocate
// a new surface, so I'll assume I'm reusing the 
// old surface as a destination one).
void IMGfToSurface(SDL_Surface *s, IMGf *img)
{
  RGBf *d = img->data;
  RGBu *p = (RGBu*)s->pixels;  

  for(int j = 0; j < s->h; j++) {
    for(int i = 0; i < s->w; i++, p++,d++) {
      p->r = d->r > 1.0f ? 255 : d->r * 255.0f;
      p->g = d->g > 1.0f ? 255 : d->g * 255.0f;
      p->b = d->b > 1.0f ? 255 : d->b * 255.0f;
    }
  }
}

// Treat RGB as a vector and normalize it.
// Used to calculate rgb-based distance.
void Normalize(RGBf *p) {
  float dst = sqrt(p->r * p->r + p->g * p->g + p->b * p->b);
  p->r /= dst;
  p->g /= dst;
  p->b /= dst;
}

// Treat RGB as 3D points and calculate
// standard distance between them.
// Used to calculate rgb-based distance.
float Dist(RGBf *a, RGBf *b) {
  float dr = a->r - b->r;
  float dg = a->g - b->g;
  float db = a->b - b->b;

  return sqrtf(dr*dr + dg*dg + db*db);
}

// Calculate grayscale value from RGB by calculating
// avg of colors.
float Light(RGBf *p) {
  return (p->r + p->g + p->b) / 3.0f;
}

// Colorize the pixel. Power is 0.0 to ~0.5 (I actually
// assume it, I didn't check), so I multiply by 2 here to
// get 0.0 to 1.0.
void Colorize(RGBf *p, RGBf *c, float power) {
  power *= 2;
  p->r = c->r * power;
  p->g = c->g * power;
  p->b = c->b * power;
}

// Calculate max of 3 floats.
inline float Max3(float a, float b, float c, int *which) {
  float M = a; *which = 0;
  if(b > a) { M = b; *which = 1; }
  if(c > M) { M = c; *which = 2; }
  return M;
}

// Calculate min of 3 floats.
inline float Min3(float a, float b, float c) {
  float m = a;
  if(b < a) m = b;
  if(c < m) m = c;
  return m;
}

// A lame version of floating point modulo.
// Actually could have used FPREM x86 instruction or the
// fmod() from math, but.. for got it exists and I already
// have this coded, so... :)
float Modf(float what, float mod) {
  while(what > mod) what -= mod;
  while(what < 0)   what += mod;
  return what;
}

// Get Hue from RGB, according to formula from
// wikipedia:
// http://en.wikipedia.org/wiki/HSL_and_HSV#Hue_and_chroma
float RGBftoHue(RGBf *p) {
  int which_M = 0;
  float M = Max3(p->r, p->g, p->b, &which_M);
  float m = Min3(p->r, p->g, p->b);
  float C = M - m;

  float Hp = 0.0f;
  if(C == 0.0f) return Hp;

  switch(which_M) {
    case 0: Hp = (p->g - p->b) / C; break;
    case 1: Hp = (p->b - p->r) / C + 2; break;
    case 2: Hp = (p->r - p->g) / C + 4; break;
  }

  float H = 60.0f * Modf(Hp, 6.0f);
  return H;
}

// A lame version of angular distance on a circle
// between two hues.
float DistHue(RGBf *a, RGBf *b) {
  float d1 = fabs(RGBftoHue(a) - RGBftoHue(b));
  float d2 = fabs((RGBftoHue(a) + 360) - RGBftoHue(b));
  float d3 = fabs((RGBftoHue(a) - 360) - RGBftoHue(b));
  return Min3(d1,d2,d3);
}

// Make the look as cool as perihelion did (well OK,
// almost as cool ;>).
// Currently using hue-based distance method.
void CoolTheImg(IMGf *img)
{
  RGBf *p = img->data;

  if(g_UseHueDist) {
    // HUE
    for(int j = 0; j < img->h; j++) {
      for(int i = 0; i < img->w; i++ ,p++) {
        RGBf c = *p;

        float d_sepia = DistHue(&c, &g_Sepia);

        float tr = g_BayerMatrix[j % 4][i % 4];
        float l  = Light(p);

        float power = l + tr;

        power = float(int(power * 16)) / 16.0f;

        if(d_sepia < (g_Angle / 2.0f))
          Colorize(p, &g_SepiaNorm, power);
        else
          Colorize(p, &g_GrayNorm, power);
      }
    }
  } else {
    // RGB
    for(int j = 0; j < img->h; j++) {
      for(int i = 0; i < img->w; i++ ,p++) {
        RGBf c = *p;

        Normalize(&c);

        float d = Dist(&c, &g_SepiaNorm);

        float tr = g_BayerMatrix[j % 4][i % 4];
        float l  = Light(p);

        float power = l + tr;

        power = float(int(power * 16)) / 16.0f;

        if(d < g_Angle)
          Colorize(p, &g_SepiaNorm, power);
        else
          Colorize(p, &g_GrayNorm, power);
      }
    }
  }
}

// Show usage
void Usage() {
  puts("usage: 32colors --hue <infile.png> <angle> [<r> <g> <b>]\n"
       "       32colors --rgb <infile.png> <dist> [<r> <g> <b>]\n"
       "Default <r> <g> <b> is sepia/orange color.\n"
       "<angle> is from 0 to 360\n"
       "<dist> is from 0.0 to 1.0");
}



// Main.
int
main(int argc, char **argv)
{
  // Banner
  puts("32colors by gynvael.coldwind//vx\n"
       "Inspired by the old Perihelion game for Amiga.");

  // Args?
  if(argc != 4 && argc != 7) {
    Usage();
    return 1;
  }

  if(strcmp(argv[1], "--hue") == 0) {
    g_UseHueDist = true;
  } else if(strcmp(argv[1], "--rgb") == 0) {
    g_UseHueDist = false;
  } else {
    Usage();
    return 1;
  }

  // Get args.
  g_Angle = atof(argv[3]);

  if(argc == 7) {
    g_Sepia.r = FF(atoi(argv[4]));
    g_Sepia.g = FF(atoi(argv[5]));
    g_Sepia.b = FF(atoi(argv[6]));
    g_SepiaNorm = g_Sepia;
  }

  // Normalize.
  Normalize(&g_SepiaNorm);
  Normalize(&g_GrayNorm);

  // Load image.
  SDL_Surface *src = IMG_Load(argv[2]);
  if(src == NULL) {
    printf("error: could not load image \"%s\"\n", argv[2]);
    exit(2);
  }

  // Some info.
  printf("Primary color  : %.2f, %.2f, %.2f\n", 
      g_Sepia.r, g_Sepia.g, g_Sepia.b);
  printf("Primary angle  : %i\n", (int)g_Angle);

  printf("Secondary color: %.2f, %.2f, %.2f\n", 
      g_Gray.r, g_Gray.g, g_Gray.b);


  // Do the magic.
  IMGf *img = SurfaceToIMGf(src);
  CoolTheImg(img);
  IMGfToSurface(src, img);

  // Save the image.
  char d[256];
  snprintf(d, sizeof(d)-1, "%s_%.3u.bmp", argv[2], (unsigned int)g_Angle);
  SDL_SaveBMP(src, d);

  // Done.
  printf("Done (%s written).\n", d);
  return 0;
}


